from typing import Tuple

from velour import Annotation, Datum, Label, Prediction, enums
from velour.client import Client, ClientException, Model
from velour.schemas import BoundingBox


def _parse_chariot_predict_image_classification(
    datum: Datum,
    labels: dict,
    result: list,
    label_key: str = "class_label",
):
    # validate result
    if not isinstance(result, list):
        raise TypeError("expected format List[str]")
    if len(result) != 1:
        raise ValueError("cannot have more than one result per datum")
    if not isinstance(result[0], str):
        raise TypeError("expected format List[str]")

    # create prediction
    return Prediction(
        datum=datum,
        annotations=[
            Annotation(
                task_type=enums.TaskType.CLASSIFICATION,
                labels=[
                    Label(
                        key=label_key,
                        value=label,
                        score=1.0 if label == result[0] else 0.0,
                    )
                    for label in labels
                ],
            )
        ],
    )


def _parse_chariot_predict_proba_image_classification(
    datum: Datum,
    labels: dict,
    result: list,
    label_key: str = "class_label",
):
    # validate result
    if not isinstance(result, list):
        raise TypeError("expected format List[List[float]]")
    if len(result) != 1:
        raise ValueError("cannot have more than one result per datum")
    if not isinstance(result[0], list):
        raise TypeError("expected format List[List[float]]")
    if len(labels) != len(result[0]):
        raise ValueError("number of labels does not equal number of scores")

    # create prediction
    labels = {v: k for k, v in labels.items()}
    return Prediction(
        datum=datum,
        annotations=[
            Annotation(
                task_type=enums.TaskType.CLASSIFICATION,
                labels=[
                    Label(key=label_key, value=labels[i], score=score)
                    for i, score in enumerate(result[0])
                ],
            )
        ],
    )


def _parse_chariot_detect_image_object_detection(
    datum: Datum,
    result: dict,
    label_key: str = "class_label",
):
    # validate result
    if not isinstance(result, list):
        raise TypeError("image object detection expected format List[Dict[]]")
    if len(result) != 1:
        raise ValueError("cannot have more than one result per datum")
    if not isinstance(result[0], dict):
        raise TypeError("image object detection expected format List[Dict[]]")
    result = result[0]

    # validate result
    expected_keys = {
        "num_detections",
        "detection_classes",
        "detection_boxes",
        "detection_scores",
    }
    if set(result.keys()) != expected_keys:
        raise ValueError(
            f"Expected `dets` to have keys {expected_keys} but got {result.keys()}"
        )

    # create prediction
    return Prediction(
        datum=datum,
        annotations=[
            Annotation(
                task_type=enums.TaskType.DETECTION,
                labels=[Label(key=label_key, value=label, score=float(score))],
                bounding_box=BoundingBox.from_extrema(
                    ymin=box[0], xmin=box[1], ymax=box[2], xmax=box[3]
                ),
            )
            for box, score, label in zip(
                result["detection_boxes"],
                result["detection_scores"],
                result["detection_classes"],
            )
        ],
    )


def create_model_from_chariot(
    client: Client,
    model,
):
    return Model.create(
        client=client,
        name=model.id,
        metadata={
            "integration": "chariot",
            "title": model.name,
            "description": model._meta.summary,
            "project_id": model.project_id,
        },
    )


def get_prediction_parser_from_chariot(
    task_type: str, action: str, label_key: str, class_labels: list = None
):
    if task_type == "Image Classification":
        if action == "predict":

            def velour_parser(datum: Datum, result):
                return _parse_chariot_predict_image_classification(
                    datum, class_labels, result=result, label_key=label_key
                )

        elif action == "predict_proba":

            def velour_parser(datum: Datum, result):
                return _parse_chariot_predict_proba_image_classification(
                    datum, class_labels, result=result, label_key=label_key
                )

        else:
            raise NotImplementedError(
                f"action `{action}` not supported for task type `Image Classification`"
            )

    elif task_type == "Object Detection":

        def velour_parser(datum: Datum, result):
            return _parse_chariot_detect_image_object_detection(
                datum=datum,
                result=result,
                label_key=label_key,
            )

    elif task_type == "Image Segmentation":
        raise NotImplementedError(
            "Image Segmentation has not been implemented."
        )
    else:
        raise NotImplementedError(task_type)

    return velour_parser


def get_chariot_model_integration(
    client: Client, model, action: str, label_key: str = "class_label"
) -> Tuple[Model, callable]:
    """Creates or retrieves Velour model and parsing function for annotation type.

    Returns
    ----------
    tuple
        model: velour.client.Model
        parser: callable
            datum: velour.schemas.Datum
            result: chariot result
    """

    # get or create model
    try:
        velour_model = Model.get(client, model.id)
    except ClientException as e:
        if "does not exist" not in str(e):
            raise e
        velour_model = create_model_from_chariot(client, model)

    # retrieve task-related parser
    velour_parser = get_prediction_parser_from_chariot(
        task_type=model.task.value,
        action=action,
        label_key=label_key,
        class_labels=model.class_labels,
    )

    return (velour_model, velour_parser)
